"""Tests for WolkAboutProtocolMessageFactory."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import sys
import time
import unittest

sys.path.append("..")  # noqa

from wolk.model.state import State
from wolk.model.actuator_status import ActuatorStatus
from wolk.model.alarm import Alarm
from wolk.model.message import Message
from wolk.model.sensor_reading import SensorReading
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType
from wolk.model.firmware_update_error_type import FirmwareUpdateErrorType
from wolk.wolkabout_protocol_message_factory import (
    WolkAboutProtocolMessageFactory as WAPMF,
)

unittest.util._MAX_LENGTH = 2000


class WolkAboutProtocolMessageFactoryTests(unittest.TestCase):
    """Tests for serializing messages using WolkAbout Protocol."""

    def test_init(self):
        """Test that object is created with correct device key."""
        device_key = "some_key"

        factory = WAPMF(device_key)

        self.assertEqual(device_key, factory.device_key)

    def test_sensor_bool(self):
        """Test that valid message is created for bool sensor reading."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "B"
        value = False
        expected_value = str(value).lower()
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.SENSOR_READING
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            [{"data": expected_value, "utc": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        reading = SensorReading(reference, value, timestamp)

        serialized_message = factory.make_from_sensor_reading(reading)

        self.assertEqual(expected_message, serialized_message)

    def test_sensor_string_with_newline(self):
        """Test valid message for string with newline sensor reading."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "SNL"
        value = "string\nin\na\ncouple\nof\nrows"
        expected_value = value  # escaped in json.dumps
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.SENSOR_READING
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            [{"data": expected_value, "utc": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        reading = SensorReading(reference, value, timestamp)

        serialized_message = factory.make_from_sensor_reading(reading)

        self.assertEqual(expected_message, serialized_message)

    def test_sensor_with_timestamp(self):
        """Test valid message for sensor reading with timestamp."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "TS"
        value = "value"
        timestamp = int(round(time.time() * 1000))

        reading = SensorReading(reference, value, timestamp)

        serialized_message = factory.make_from_sensor_reading(reading)

        self.assertIn("utc", json.loads(serialized_message.payload)[0])

    def test_sensor_multi_value_float(self):
        """Test valid message for multi-value float sensor reading."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "MVF"
        value = (12.3, 45.6)
        expected_value = ",".join(map(str, value))
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.SENSOR_READING
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            [{"data": expected_value, "utc": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        reading = SensorReading(reference, value, timestamp)

        serialized_message = factory.make_from_sensor_reading(reading)

        self.assertEqual(expected_message, serialized_message)

    def test_sensor_multi_value_string(self):
        """Test valid message for multi-value string sensor reading."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "MVS"
        value = ("string1", "string2")
        expected_value = ",".join(value)
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.SENSOR_READING
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            [{"data": expected_value, "utc": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        reading = SensorReading(reference, value, timestamp)

        serialized_message = factory.make_from_sensor_reading(reading)

        self.assertEqual(expected_message, serialized_message)

    def test_sensor_multi_value_string_with_newline(self):
        """Test message for multi-value string with newline sensor reading."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "MVSNL"
        value = ("string1\nstring1", "string2")
        expected_value = ",".join(value)  # new line escaped in json dumps
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.SENSOR_READING
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            [{"data": expected_value, "utc": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        reading = SensorReading(reference, value, timestamp)

        serialized_message = factory.make_from_sensor_reading(reading)

        self.assertEqual(expected_message, serialized_message)

    def test_multiple_sensor_readings(self):
        """Test message for multiple sensor readings."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference1 = "B"
        value1 = False
        reference2 = "T"
        value2 = (12, 23, 34)
        reference3 = "F"
        value3 = 45.6
        readings = {reference1: value1, reference2: value2, reference3: value3}
        timestamp = int(round(time.time() * 1000))

        expected_topic = (
            WAPMF.SENSOR_READING + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            {
                reference1: str(value1).lower(),
                reference2: ",".join(map(str, value2)),
                reference3: str(value3),
                "utc": int(timestamp),
            }
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_sensor_readings(
            readings, timestamp
        )

        self.assertEqual(expected_message, serialized_message)

    def test_multiple_sensor_readings_historical(self):
        """Test message for multiple sensor readings."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "HS"
        value_1 = False
        timestamp_1 = int(round(time.time() * 1000))
        value_2 = (12, 23, 34)
        timestamp_2 = timestamp_1 - 600000
        value_3 = 45.6
        timestamp_3 = timestamp_2 - 600000

        reading_1 = SensorReading(reference, value_1, timestamp_1)
        reading_2 = SensorReading(reference, value_2, timestamp_2)
        reading_3 = SensorReading(reference, value_3, timestamp_3)
        readings = [reading_1, reading_2, reading_3]

        expected_topic = (
            WAPMF.SENSOR_READING
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            [
                {"data": str(value_1).lower(), "utc": timestamp_1},
                {"data": ",".join(map(str, value_2)), "utc": timestamp_2},
                {"data": str(value_3), "utc": timestamp_3},
            ]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_sensor_reading(readings)

        self.assertEqual(expected_message, serialized_message)

    def test_multiple_sensor_readings_no_timestamp(self):
        """Test message for multiple sensor readings with no user timestamp."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference1 = "B"
        value1 = False
        reference2 = "T"
        value2 = (12, 23, 34)
        reference3 = "F"
        value3 = 45.6
        readings = {reference1: value1, reference2: value2, reference3: value3}
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.SENSOR_READING + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            {
                reference1: str(value1).lower(),
                reference2: ",".join(map(str, value2)),
                reference3: str(value3),
                "utc": timestamp,
            }
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_sensor_readings(readings)

        self.assertEqual(expected_message, serialized_message)

    def test_alarm_no_timestamp(self):
        """Test message for alarm without user defined timestamp."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "A"
        value = False
        expected_value = str(value).lower()
        timestamp = round(time.time()) * 1000

        expected_topic = (
            WAPMF.ALARM
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            {"active": expected_value, "utc": timestamp}
        )
        expected_message = Message(expected_topic, expected_payload)

        alarm = Alarm(reference, value)

        serialized_message = factory.make_from_alarm(alarm)

        self.assertEqual(expected_message, serialized_message)

    def test_alarm_with_timestamp(self):
        """Test message for alarm with timestamp."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "A"
        value = True
        expected_value = str(value).lower()
        timestamp = int(round(time.time() * 1000))

        expected_topic = (
            WAPMF.ALARM
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            {"active": expected_value, "utc": timestamp}
        )
        expected_message = Message(expected_topic, expected_payload)

        alarm = Alarm(reference, value, timestamp=timestamp)

        serialized_message = factory.make_from_alarm(alarm)

        self.assertEqual(expected_message, serialized_message)

    def test_actuator_bool(self):
        """Test message for bool actuator with ready state."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "A"
        value = True
        status = State.READY
        expected_value = str(value).lower()

        expected_topic = (
            WAPMF.ACTUATOR_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            {"status": status.value, "value": expected_value}
        )
        expected_message = Message(expected_topic, expected_payload)

        actuator = ActuatorStatus(reference, status, value)

        serialized_message = factory.make_from_actuator_status(actuator)

        self.assertEqual(expected_message, serialized_message)

    def test_actuator_float(self):
        """Test message for float actuator with ready state."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "A"
        value = 12.3
        status = State.READY
        expected_value = str(value)

        expected_topic = (
            WAPMF.ACTUATOR_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            {"status": status.value, "value": expected_value}
        )
        expected_message = Message(expected_topic, expected_payload)

        actuator = ActuatorStatus(reference, status, value)

        serialized_message = factory.make_from_actuator_status(actuator)

        self.assertEqual(expected_message, serialized_message)

    def test_actuator_string_with_newline(self):
        """Test message for string with newline actuator with ready state."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "A"
        value = "string\nstring"
        status = State.READY
        expected_value = value  # escaped in json.dumps

        expected_topic = (
            WAPMF.ACTUATOR_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
            + WAPMF.CHANNEL_DELIMITER
            + WAPMF.REFERENCE_PATH_PREFIX
            + reference
        )
        expected_payload = json.dumps(
            {"status": status.value, "value": expected_value}
        )
        expected_message = Message(expected_topic, expected_payload)

        actuator = ActuatorStatus(reference, status, value)

        serialized_message = factory.make_from_actuator_status(actuator)

        self.assertEqual(expected_message, serialized_message)

    def test_configuration_bool(self):
        """Test message for configuration bool."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "B"
        value = True
        configuration = {reference: value}

        expected_value = configuration.copy()
        expected_value[reference] = str(expected_value[reference]).lower()

        expected_topic = (
            WAPMF.CONFIGURATION_STATUS + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps({"values": expected_value})
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_configuration(configuration)

        self.assertEqual(expected_message, serialized_message)

    def test_configuration_float(self):
        """Test message for configuration float."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "F"
        value = 12.3
        configuration = {reference: value}

        expected_value = configuration.copy()
        expected_value[reference] = str(expected_value[reference]).lower()

        expected_topic = (
            WAPMF.CONFIGURATION_STATUS + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps({"values": expected_value})
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_configuration(configuration)

        self.assertEqual(expected_message, serialized_message)

    def test_configuration_string_with_newline(self):
        """Test message for configuration string with newline."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        reference = "SNL"
        value = "string\nstring"
        configuration = {reference: value}

        expected_value = configuration.copy()
        expected_value[reference] = str(expected_value[reference]).lower()

        expected_topic = (
            WAPMF.CONFIGURATION_STATUS + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps({"values": expected_value})
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_configuration(configuration)

        self.assertEqual(expected_message, serialized_message)

    def test_file_list_update(self):
        """Test message for file list update."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_list = ["file1.txt", "file2.bin", "file3 with spaces.jpg"]

        expected_topic = (
            WAPMF.FILE_LIST_UPDATE + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            [{"fileName": file} for file in file_list]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_file_list_update(file_list)

        self.assertEqual(expected_message, serialized_message)

    def test_file_list_request(self):
        """Test message for file list request."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_list = ["file1.txt", "file2.bin", "file3 with spaces.jpg"]

        expected_topic = (
            WAPMF.FILE_LIST_RESPONSE + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            [{"fileName": file} for file in file_list]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = factory.make_from_file_list_request(file_list)

        self.assertEqual(expected_message, serialized_message)

    def test_file_package_request(self):
        """Test message file package request."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_name = "file_name"
        chunk_index = 0
        chunk_size = 256

        expected_topic = (
            WAPMF.FILE_BINARY_REQUEST + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            {
                "fileName": file_name,
                "chunkIndex": chunk_index,
                "chunkSize": chunk_size,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_package_request(
            file_name, chunk_index, chunk_size
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_status_ready(self):
        """Test message for file management status with file ready."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_name = "file_name"
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        expected_topic = (
            WAPMF.FILE_UPLOAD_STATUS + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            {"fileName": file_name, "status": status.status.value}
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_file_management_status(
            status, file_name
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_status_error(self):
        """Test message for file management status with error status."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_name = "file_name"
        status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.UNSPECIFIED_ERROR,
        )
        expected_topic = (
            WAPMF.FILE_UPLOAD_STATUS + WAPMF.DEVICE_PATH_PREFIX + device_key
        )
        expected_payload = json.dumps(
            {
                "fileName": file_name,
                "status": status.status.value,
                "error": status.error.value,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_file_management_status(
            status, file_name
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_url_status_ready(self):
        """Test message for file url management status with file ready."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_name = "file_name"
        file_url = "file_url"
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        expected_topic = (
            WAPMF.FILE_URL_DOWNLOAD_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
        )
        expected_payload = json.dumps(
            {
                "fileUrl": file_url,
                "status": status.status.value,
                "fileName": file_name,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_file_url_status(
            file_url, status, file_name
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_url_status_error(self):
        """Test message for file url management status with error status."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        file_url = "file_url"
        status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.MALFORMED_URL,
        )
        expected_topic = (
            WAPMF.FILE_URL_DOWNLOAD_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
        )
        expected_payload = json.dumps(
            {
                "fileUrl": file_url,
                "status": status.status.value,
                "error": status.error.value,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_file_url_status(
            file_url, status
        )

        self.assertEqual(expected_message, serialized_message)

    def test_firmware_update_status(self):
        """Test message for firmware update status."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        expected_topic = (
            WAPMF.FIRMWARE_UPDATE_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
        )
        expected_payload = json.dumps({"status": status.status.value})
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_firmware_update_status(status)

        self.assertEqual(expected_message, serialized_message)

    def test_firmware_update_status_error(self):
        """Test message for firmware update status."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.ERROR,
            FirmwareUpdateErrorType.INSTALLATION_FAILED,
        )
        expected_topic = (
            WAPMF.FIRMWARE_UPDATE_STATUS
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
        )
        expected_payload = json.dumps(
            {"status": status.status.value, "error": status.error.value}
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_firmware_update_status(status)

        self.assertEqual(expected_message, serialized_message)

    def test_last_will_message(self):
        """Test message for last will."""
        device_key = "some_key"
        factory = WAPMF(device_key)

        expected_topic = WAPMF.LAST_WILL + device_key
        expected_payload = None
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_last_will_message()

        self.assertEqual(expected_message, serialized_message)

    def test_keep_alive_message(self):
        """Test message for keep alive."""
        device_key = "some_key"
        factory = WAPMF(device_key)

        expected_topic = WAPMF.KEEP_ALIVE + device_key
        expected_payload = None
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_keep_alive_message()

        self.assertEqual(expected_message, serialized_message)

    def test_firmware_version_update(self):
        """Test message for firmware version update."""
        device_key = "some_key"
        factory = WAPMF(device_key)
        version = "v1.0.0"

        expected_topic = (
            WAPMF.FIRMWARE_VERSION_UPDATE
            + WAPMF.DEVICE_PATH_PREFIX
            + device_key
        )
        expected_payload = version
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = factory.make_from_firmware_version_update(version)

        self.assertEqual(expected_message, serialized_message)



"""Heart and soul of this library. Wraps in all functionality."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from inspect import signature
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk import logger_factory
from wolk.interface.connectivity_service import ConnectivityService
from wolk.interface.file_management import FileManagement
from wolk.interface.firmware_handler import FirmwareHandler
from wolk.interface.firmware_update import FirmwareUpdate
from wolk.interface.message_deserializer import MessageDeserializer
from wolk.interface.message_factory import MessageFactory
from wolk.interface.message_queue import MessageQueue
from wolk.message_deque import MessageDeque
from wolk.model.actuator_status import ActuatorStatus
from wolk.model.alarm import Alarm
from wolk.model.device import Device
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.firmware_update_error_type import FirmwareUpdateErrorType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType
from wolk.model.message import Message
from wolk.model.sensor_reading import SensorReading
from wolk.model.state import State
from wolk.mqtt_connectivity_service import MQTTConnectivityService as MQTTCS
from wolk.os_file_management import OSFileManagement
from wolk.os_firmware_update import OSFirmwareUpdate
from wolk.repeating_timer import RepeatingTimer
from wolk.wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer as WAPMDeserializer,
)
from wolk.wolkabout_protocol_message_factory import (
    WolkAboutProtocolMessageFactory as WAPMFactory,
)

ConfigurationValue = Union[
    bool,
    int,
    float,
    str,
]

ActuatorValue = Tuple[State, Optional[Union[bool, int, float, str]]]

ReadingValue = Union[
    bool,
    int,
    Tuple[int, ...],
    float,
    Tuple[float, ...],
    str,
]


class WolkConnect:
    """
    Exchange data with WolkAbout IoT Platform.

    :ivar actuation_handler: Function for handling actuation commands
    :vartype actuation_handler: Callable[[str, Union[bool, int, float, str]], None] or None
    :ivar actuator_status_provider: Function for getting current actuator status
    :vartype actuator_status_provider: Callable[[str], ActuatorValue] or None
    :ivar configuration_handler: Function for handling configuration commands
    :vartype configuration_handler: Callable[[dict], None] or None
    :ivar configuration_provider: Function for getting current configuration options
    :vartype configuration_provider: Callable[[None], dict] or None
    :ivar connectivity_service: Means of sending/receiving data
    :vartype connectivity_service: ConnectivityService
    :ivar device: Contains device key and password, and actuator references
    :vartype device: Device
    :ivar file_management: File management module
    :vartype file_management: FileManagement or None
    :ivar firmware_update: Firmware update handler
    :vartype firmware_update: FirmwareUpdate or None
    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar message_deserializer: Deserializer of inbound messages
    :vartype message_deserializer: MessageDeserializer
    :ivar message_factory: Create messages to send
    :vartype message_factory: MessageFactory
    :ivar message_queue: Store data before sending
    :vartype message_queue: MessageQueue
    :ivar keep_alive_service_enabled: Enable keep alive service
    :vartype keep_alive_service_enabled: bool
    :ivar keep_alive_interval: Interval in seconds when to publish message
    :vartype keep_alive_interval: int
    :ivar keep_alive_service: Keep alive service
    :vartype keep_alive_service: RepeatingTimer or None
    """

    def __init__(
        self,
        device: Device,
        host: Optional[str] = None,
        port: Optional[int] = None,
        ca_cert: Optional[str] = None,
    ):
        """
        Provide communication with WolkAbout IoT Platform.

        :param device: Contains key and password, and actuator references
        :type device: Device
        :param host: Host name or IP address of the remote broker
        :type host: str, optional
        :param port: Network port of the server host to connect to
        :type port: int, optional
        :param ca_cert: String path to Certificate Authority certificate file
        :type ca_cert: str, optional
        """
        logger_factory.logger_factory.set_device_key(device.key)
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            f"Device: {device} ; "
            f"Host: {host} ; Port: {port} ; ca_cert: {ca_cert}"
        )

        self.device = device
        self.actuation_handler: Optional[
            Callable[[str, Union[bool, int, float, str]], None]
        ] = None
        self.actuator_status_provider: Optional[
            Callable[[str], ActuatorValue]
        ] = None

        self.configuration_handler: Optional[
            Callable[[Dict[str, ConfigurationValue]], None]
        ] = None
        self.configuration_provider: Optional[
            Callable[[None], Dict[str, ConfigurationValue]]
        ] = None
        self.file_management: Optional[FileManagement] = None
        self.firmware_update: Optional[FirmwareUpdate] = None
        self.message_queue: MessageQueue = MessageDeque()
        self.message_factory: MessageFactory = WAPMFactory(device.key)
        self.message_deserializer: MessageDeserializer = WAPMDeserializer(
            self.device
        )

        wolk_ca_cert = os.path.join(os.path.dirname(__file__), "ca.crt")
        last_will_message = self.message_factory.make_last_will_message()

        if host and port and ca_cert:
            self.connectivity_service: ConnectivityService = MQTTCS(
                device,
                self.message_deserializer.get_inbound_topics(),
                last_will_message,
                host=host,
                port=int(port),
                ca_cert=ca_cert,
            )
        elif host and port:
            self.connectivity_service = MQTTCS(
                device,
                self.message_deserializer.get_inbound_topics(),
                last_will_message,
                host=host,
                port=int(port),
            )
        else:
            self.connectivity_service = MQTTCS(
                device,
                self.message_deserializer.get_inbound_topics(),
                last_will_message,
                ca_cert=wolk_ca_cert,
            )

        self.connectivity_service.set_inbound_message_listener(
            self._on_inbound_message
        )

        self.keep_alive_service_enabled = True
        self.keep_alive_interval = 60
        self.keep_alive_service: Optional[RepeatingTimer] = None

        self.last_platform_timestamp: Optional[int] = None

    def with_actuators(  # type: ignore
        self,
        actuation_handler: Callable[[str, Union[bool, int, float, str]], None],
        actuator_status_provider: Callable[[str], ActuatorValue],
    ):
        """
        Enable controlling actuators on the device from the Platform.

        :param actuation_handler: Handle actuation commands
        :type actuation_handler: Callable[[str, Union[bool, int, float, str]], None]
        :param actuator_status_provider: Read actuator status
        :type actuator_status_provider: Callable[[str], ActuatorValue]
        """
        self.logger.debug(
            f"Actuation handler: {actuation_handler} ; "
            f"Actuator status provider: {actuator_status_provider}"
        )

        if not callable(actuation_handler):
            raise ValueError(f"{actuation_handler} is not a callable!")
        if len(signature(actuation_handler).parameters) != 2:
            raise ValueError(f"{actuation_handler} invalid signature!")
        self.actuation_handler = actuation_handler

        if not callable(actuator_status_provider):
            raise ValueError(f"{actuator_status_provider} is not a callable!")
        if len(signature(actuator_status_provider).parameters) != 1:
            raise ValueError(f"{actuator_status_provider} invalid signature!")
        self.actuator_status_provider = actuator_status_provider

        return self

    def with_configuration(  # type: ignore
        self,
        configuration_handler: Callable[[Dict[str, ConfigurationValue]], None],
        configuration_provider: Callable[
            [None], Dict[str, ConfigurationValue]
        ],
    ):
        """
        Enable setting device's configuration options from the Platform.

        :param configuration_handler: Handle configuration commands
        :type configuration_handler: Callable[[Dict[str,ConfigurationValue]], None]
        :param configuration_provider: Read configuration options
        :type configuration_provider: Callable[[None], Dict[str, ConfigurationValue]]
        """
        self.logger.debug(
            f"Configuration handler: {configuration_handler} ; "
            f"Configuration provider: {configuration_provider}"
        )

        if not callable(configuration_handler):
            raise ValueError(f"{configuration_handler} is not a callable!")
        if len(signature(configuration_handler).parameters) != 1:
            raise ValueError(f"{configuration_handler} invalid signature!")
        self.configuration_handler = configuration_handler

        if not callable(configuration_provider):
            raise ValueError(f"{configuration_provider} is not a callable!")
        if len(signature(configuration_provider).parameters) != 0:
            raise ValueError(f"{configuration_provider} invalid signature!")
        self.configuration_provider = configuration_provider

        return self

    def with_file_management(  # type: ignore
        self,
        preferred_package_size: int,
        max_file_size: int,
        file_directory: str,
    ):
        """
        Enable file management on the device.

        :param preferred_package_size: Size of file package chunk in bytes
        :type preferred_package_size: int
        :param max_file_size: Maximum supported file size in bytes
        :type max_file_size: int
        :param file_directory: Directory where files are stored
        :type file_directory: str
        """
        self.logger.debug(
            f"Preferred package size: {preferred_package_size}, "
            f"maximum file size: {max_file_size}, "
            f"file directory: '{file_directory}'"
        )

        self.file_management = OSFileManagement(
            self._on_file_upload_status,
            self._on_package_request,
            self._on_file_url_status,
        )
        self.file_management.configure(
            preferred_package_size,
            max_file_size,
            file_directory,
        )

        return self

    def with_firmware_update(self, firmware_handler: FirmwareHandler):  # type: ignore
        """
        Enable firmware update for device.

        Requires that file management is previously enabled on device.

        :param firmware_handler: Provide firmware version and handle installation
        :type firmware_handler: FirmwareHandler
        """
        self.logger.debug(f"Firmware handler: {firmware_handler}")
        if self.file_management is None:
            raise RuntimeError(
                "File management must be enabled before firmware update"
            )
        self.firmware_update = OSFirmwareUpdate(
            firmware_handler, self._on_firmware_update_status
        )

        message = self.message_factory.make_from_firmware_version_update(
            self.firmware_update.get_current_version()
        )
        self.message_queue.put(message)
        self.firmware_update.report_result()

        return self

    def with_custom_message_queue(self, message_queue: MessageQueue):  # type: ignore
        """
        Set custom means of storing serialized messages.

        :param message_queue: Custom message queue
        :type message_queue: MessageQueue
        """
        self.logger.debug(f"Message queue: {message_queue}")
        if not isinstance(message_queue, MessageQueue):
            raise ValueError(
                "Provided message queue does not implement MessageQueue"
            )

        self.message_queue = message_queue

        return self

    def with_custom_protocol(  # type: ignore
        self,
        message_factory: MessageFactory,
        message_deserializer: MessageDeserializer,
    ):
        """
        Provide a custom protocol to use for communication with the Platform.

        :param message_factory: Creator of messages to be sent to the Platform
        :type message_factory: MessageFactory
        :param message_deserializer: Deserializer of messages from the Platform
        :type message_deserializer: MessageDeserializer
        """
        self.logger.debug(
            f"Message factory: {message_factory} ; "
            f"message deserializer: {message_deserializer}"
        )
        if not isinstance(message_factory, MessageFactory):
            raise ValueError("Invalid message factory provided")
        self.message_factory = message_factory

        if not isinstance(message_deserializer, MessageDeserializer):
            raise ValueError("Invalid message deserializer provided")
        self.message_deserializer = message_deserializer

        return self

    def with_custom_connectivity(  # type: ignore
        self, connectivity_service: ConnectivityService
    ):
        """
        Provide a custom way to communicate with the Platform.

        :param connectivity_service: Custom connectivity service
        :type connectivity_service: ConnectivityService
        """
        self.logger.debug(f"Connectivity service: {connectivity_service}")
        if not isinstance(connectivity_service, ConnectivityService):
            raise ValueError("Invalid connectivity service provided")
        self.connectivity_service = connectivity_service
        self.connectivity_service.set_inbound_message_listener(
            self._on_inbound_message
        )

        return self

    def with_keep_alive_service(  # type: ignore
        self, enabled: bool, interval: Optional[int] = None
    ):
        """
        Enable or disable keep alive service.

        :param enabled: Enable or disable keep alive service
        :type enabled: bool
        :param interval: Interval in seconds, default is 60
        :type interval: int or None
        """
        if not enabled:
            self.keep_alive_service_enabled = False
            self.keep_alive_service = None

        if interval:
            self.keep_alive_interval = interval

        return self

    def _send_keep_alive(self) -> None:
        if not self.connectivity_service.is_connected():
            return
        message = self.message_factory.make_keep_alive_message()
        self.connectivity_service.publish(message)

    def connect(self) -> None:
        """
        Connect the device to the WolkAbout IoT Platform.

        If the connection is made, then it also sends information
        regarding current actuator statuses, configuration option values,
        list of files present on device, current firmware version
        and the result of the firmware update process.
        """
        self.logger.debug("Connecting")

        if self.connectivity_service.is_connected():
            self.logger.info("Already connected")
            return

        try:
            self.connectivity_service.connect()
        except Exception as exception:
            self.logger.exception(
                f"Something went wrong when trying to connect: {exception}"
            )
            return

        if self.connectivity_service.is_connected():

            if self.actuator_status_provider:
                for reference in self.device.actuator_references:
                    self.publish_actuator_status(reference)
            if self.configuration_provider:
                self.publish_configuration()
            if self.file_management:
                file_list = self.file_management.get_file_list()
                message = self.message_factory.make_from_file_list_update(
                    file_list
                )
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
            if self.firmware_update:
                message = (
                    self.message_factory.make_from_firmware_version_update(
                        self.firmware_update.get_current_version()
                    )
                )
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)

                self.firmware_update.report_result()

            if self.keep_alive_service_enabled:
                self._send_keep_alive()
                self.keep_alive_service = RepeatingTimer(
                    self.keep_alive_interval, self._send_keep_alive
                )
                self.keep_alive_service.start()

    def disconnect(self) -> None:
        """Disconnect the device from WolkAbout IoT Platform."""
        if not self.connectivity_service.is_connected():
            return
        self.logger.debug("Disconnecting")
        if self.keep_alive_service is not None:
            self.keep_alive_service.cancel()
        self.connectivity_service.disconnect()

    def add_sensor_reading(
        self,
        reference: str,
        value: Union[ReadingValue, List[Tuple[ReadingValue, int]]],
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Place a sensor reading into storage.

        Pass a single reading or multiple readings as a list
        of tuples where the first element is the reading data
        and the second element is timestamp when the reading occurred.
        If passing a list of readings, omit the `timestamp` argument.

        :param reference: The reference of the sensor
        :type reference: str
        :param value: The value of the sensor reading
        :type value: Union[ReadingValue, List[Tuple[ReadingValue, int]]]
        :param timestamp: Unix timestamp. If not provided, library will assign
        :type timestamp: Optional[int]
        """
        self.logger.debug(
            f"Adding sensor reading: reference = '{reference}', "
            f"value = {value}, timestamp = {timestamp}"
        )
        if isinstance(value, list):
            reading: Union[SensorReading, List[SensorReading]] = [
                SensorReading(reference, data[0], data[1]) for data in value
            ]
        else:
            reading = SensorReading(reference, value, timestamp)
        message = self.message_factory.make_from_sensor_reading(reading)
        self.message_queue.put(message)

    def add_sensor_readings(
        self,
        readings: Dict[str, ReadingValue],
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Place multiple sensor readings into storage.

        :param readings: Dictionary of reference: value pairs
        :type readings: Dict[str, Union[bool, int, Tuple[int, ...] float, Tuple[float, ...], str]]
        :param timestamp: Unix timestamp. If not provided, library will assign
        :type timestamp: Optional[int]
        """
        self.logger.debug(
            f"Adding sensor readings: readings:{readings}, "
            f"timestamp = {timestamp}"
        )
        message = self.message_factory.make_from_sensor_readings(
            readings, timestamp
        )
        self.message_queue.put(message)

    def add_alarm(
        self,
        reference: str,
        active: bool,
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Publish an alarm to WolkAbout IoT Platform.

        :param reference: Reference of the alarm
        :type reference: str
        :param active: Current state of the alarm
        :type active: bool
        :param timestamp: Unix timestamp. If not provided, current time will be used
        :type timestamp: Optional[int]
        """
        self.logger.debug(
            f"Add alarm event: reference = '{reference}', "
            f"active = {active}, timestamp = {timestamp}"
        )
        alarm = Alarm(reference, active, timestamp)
        message = self.message_factory.make_from_alarm(alarm)
        self.message_queue.put(message)

    def publish(self) -> None:
        """Publish all currently stored messages to WolkAbout IoT Platform."""
        self.logger.debug("Publishing")
        if not self.connectivity_service.is_connected():
            self.logger.warning("Not connected, unable to publish messages")
            return
        while True:
            message = self.message_queue.peek()
            if message is None:
                break

            if self.connectivity_service.publish(message):
                self.message_queue.get()
            else:
                self.logger.warning(f"Failed to publish message: {message}")
                break
        self.logger.debug("Publishing ended")

    def publish_actuator_status(self, reference: str) -> None:
        """
        Publish the current actuator status to WolkAbout IoT Platform.

        :param reference: reference of the actuator
        :type reference: str
        """
        if not self.connectivity_service.is_connected():
            self.logger.warning("Not connected, unable to publish message!")
            return
        self.logger.debug(
            f"Publishing actuator status for reference '{reference}'"
        )
        if not self.actuator_status_provider:
            self.logger.error("No actuator status provider is set!")
            return

        actuator_status = self.actuator_status_provider(reference)
        if actuator_status is None:
            self.logger.error(  # type: ignore
                "Actuator status provider did not return "
                f"status for reference '{reference}'"
            )
            return
        message = self.message_factory.make_from_actuator_status(
            ActuatorStatus(reference, actuator_status[0], actuator_status[1])
        )

        if not self.connectivity_service.publish(message):
            self.logger.error(
                f"Failed to publish actuator status for reference {reference}!"
            )
            self.message_queue.put(message)

    def publish_configuration(self) -> None:
        """Publish current device configuration to WolkAbout IoT Platform."""
        if not self.connectivity_service.is_connected():
            self.logger.warning("Not connected, unable to publish message!")
            return
        self.logger.debug("Publishing configuration options")
        if not self.configuration_provider:
            self.logger.error("No configuration provider is set!")
            return

        message = self.message_factory.make_from_configuration(
            self.configuration_provider()  # type: ignore
        )

        if not self.connectivity_service.publish(message):
            self.logger.error("Failed to publish configuration options!")
            self.message_queue.put(message)

    def request_timestamp(self) -> Optional[int]:
        """
        Return last received timestamp from Platform.

        If keep alive service is disabled or device didn't connect
        at least once, then this will return None.

        :returns: UTC timestamp in milliseconds
        :rtype: int or None
        """
        if self.last_platform_timestamp is None:
            self.logger.warning("No timestamp available, returning None")
            return None

        return self.last_platform_timestamp

    def _on_inbound_message(self, message: Message) -> None:
        """
        Handle inbound messages.

        :param message: message received from the Platform
        :type message: Message
        """
        if "binary" in message.topic:
            self.logger.debug(
                f"Received message: {message.topic} , "  # type: ignore
                f"{len(message.payload)}"
            )
        else:
            self.logger.debug(f"Received message: {message}")

        file_management_messages = [
            self.message_deserializer.is_file_purge_command,
            self.message_deserializer.is_file_delete_command,
            self.message_deserializer.is_file_binary_response,
            self.message_deserializer.is_file_upload_initiate,
            self.message_deserializer.is_file_upload_abort,
            self.message_deserializer.is_file_list_request,
            self.message_deserializer.is_file_list_confirm,
            self.message_deserializer.is_file_url_initiate,
            self.message_deserializer.is_file_url_abort,
        ]

        firmware_update_messages = [
            self.message_deserializer.is_firmware_install,
            self.message_deserializer.is_firmware_abort,
        ]

        if self.message_deserializer.is_actuation_command(message):
            if not self.actuation_handler or not self.actuator_status_provider:
                self.logger.warning(
                    f"Received unexpected actuation message: {message}"
                )
                return
            actuation = self.message_deserializer.parse_actuator_command(
                message
            )
            self.actuation_handler(actuation.reference, actuation.value)
            self.publish_actuator_status(actuation.reference)

            return

        if self.message_deserializer.is_configuration_command(message):
            if (
                not self.configuration_provider
                or not self.configuration_handler
            ):
                self.logger.warning(
                    f"Received unexpected configuration message: {message}"
                )
                return
            configuration = self.message_deserializer.parse_configuration(
                message
            )
            self.configuration_handler(configuration.value)
            self.publish_configuration()
            return

        if self.message_deserializer.is_keep_alive_response(message):
            timestamp = self.message_deserializer.parse_keep_alive_response(
                message
            )
            self.logger.debug(
                "Updating last platfrom timestamp "
                f"from {self.last_platform_timestamp}"
                f" to {timestamp}"
            )
            self.last_platform_timestamp = timestamp
            return

        if any(is_message(message) for is_message in file_management_messages):
            self._on_file_management_message(message)
            return

        if any(is_message(message) for is_message in firmware_update_messages):
            self._on_firmware_message(message)
            return

        self.logger.warning(f"Received unknown message: {message}")

    def _on_file_management_message(self, message: Message) -> None:

        if not self.file_management:
            self.logger.warning(
                f"Received unexpected file management message: {message}"
            )
            status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.TRANSFER_PROTOCOL_DISABLED,
            )
            message = self.message_factory.make_from_file_management_status(
                status, ""
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_upload_initiate(message):
            name, size, fhash = self.message_deserializer.parse_file_initiate(
                message
            )
            if name != "":  # ignore invalid messages
                self.file_management.handle_upload_initiation(
                    name, size, fhash
                )
            return

        if self.message_deserializer.is_file_binary_response(message):
            package = self.message_deserializer.parse_file_binary(message)
            self.file_management.handle_file_binary_response(package)
            return

        if self.message_deserializer.is_file_upload_abort(message):
            self.file_management.handle_file_upload_abort()
            return

        if self.message_deserializer.is_file_url_abort(message):
            self.file_management.handle_file_upload_abort()
            return

        if self.message_deserializer.is_file_url_initiate(message):
            file_url = self.message_deserializer.parse_file_url(message)
            if file_url != "":  # ignore invalid messages
                self.file_management.handle_file_url_download_initiation(
                    file_url
                )
            return

        if self.message_deserializer.is_file_list_request(message):
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list_request(
                file_list
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_delete_command(message):
            file_name = self.message_deserializer.parse_file_delete_command(
                message
            )
            if file_name != "":  # ignore invalid messages
                self.file_management.handle_file_delete(file_name)
                file_list = self.file_management.get_file_list()
                message = self.message_factory.make_from_file_list_update(
                    file_list
                )
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_purge_command(message):
            self.file_management.handle_file_purge()
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list_update(
                file_list
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_list_confirm(message):
            self.file_management.handle_file_list_confirm()
            return

        self.logger.warning(f"Received unknown message: {message}")

    def _on_firmware_message(self, message: Message) -> None:
        if not self.firmware_update:
            self.logger.warning(
                f"Received unexpected firmware update message: {message}"
            )
            firmware_status = FirmwareUpdateStatus(
                FirmwareUpdateStatusType.ERROR,
                FirmwareUpdateErrorType.UNSPECIFIED_ERROR,
            )
            message = self.message_factory.make_from_firmware_update_status(
                firmware_status
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_firmware_install(message):
            file_name = self.message_deserializer.parse_firmware_install(
                message
            )
            file_path = self.file_management.get_file_path(file_name)  # type: ignore
            if not file_path:
                self.logger.error(
                    f"Specified file not found on device! Message: {message}"
                )
                firmware_status = FirmwareUpdateStatus(
                    FirmwareUpdateStatusType.ERROR,
                    FirmwareUpdateErrorType.FILE_NOT_PRESENT,
                )
                message = (
                    self.message_factory.make_from_firmware_update_status(
                        firmware_status
                    )
                )
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
                return
            self.firmware_update.handle_install(file_path)
            return

        if self.message_deserializer.is_firmware_abort(message):
            self.firmware_update.handle_abort()
            return

        self.logger.warning(f"Recevied unknown firmware message: {message}")

    def _on_package_request(
        self, file_name: str, chunk_index: int, chunk_size: int
    ) -> None:
        """
        Handle file transfer package requests.

        :param file_name: The name of the file to which the chunk belongs
        :type file_name: str
        :param chunk_index: The index of the requested chunk
        :type chunk_index: int
        :param chunk_size: The size of the requested chunk
        :type chunk_size: int
        """
        message = self.message_factory.make_from_package_request(
            file_name, chunk_index, chunk_size
        )
        if not self.connectivity_service.publish(message):
            self.message_queue.put(message)

    def _on_firmware_update_status(self, status: FirmwareUpdateStatus) -> None:
        """
        Report firmware update status to WolkAbout IoT Platform.

        :param status: The status to be reported to the Platform
        :type status: FirmwareUpdateStatus
        """
        message = self.message_factory.make_from_firmware_update_status(status)
        if self.connectivity_service.is_connected():
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
        else:
            self.message_queue.put(message)
        if (
            status.status == FirmwareUpdateStatusType.COMPLETED
            and self.firmware_update
        ):
            version = self.firmware_update.get_current_version()
            message = self.message_factory.make_from_firmware_version_update(
                version
            )
            if self.connectivity_service.is_connected():
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
            else:
                self.message_queue.put(message)

    def _on_file_upload_status(
        self, file_name: str, status: FileManagementStatus
    ) -> None:
        """
        Report file upload status to WolkAbout IoT Platform.

        :param file_name: File being transfered
        :type file_name: str
        :param status: Description of current status
        :type status: FileManagementStatus
        """
        message = self.message_factory.make_from_file_management_status(
            status, file_name
        )
        if not self.connectivity_service.publish(message):
            self.message_queue.put(message)
        if (
            status.status == FileManagementStatusType.FILE_READY
            and self.file_management
        ):
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list_update(
                file_list
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)

    def _on_file_url_status(
        self,
        file_url: str,
        status: FileManagementStatus,
        file_name: Optional[str] = None,
    ) -> None:
        """
        Report file URL transfer status to WolkAbout IoT Platform.

        :param file_url: URL from where to download file
        :type file_url: str
        :param status: Description of current status
        :type status: FileManagementStatus
        :param file_name: Name of downloaded file
        :type file_name: Optional[str]
        """
        message = self.message_factory.make_from_file_url_status(
            file_url, status, file_name
        )

        if not self.connectivity_service.publish(message):
            self.message_queue.put(message)

        if file_name and self.file_management:
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list_update(
                file_list
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)



"""Everything needed for authenticating a device on WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import List


@dataclass
class Device:
    """
    Device identified by key and password, and list of actuator references.

    :ivar key: Device's key
    :vartype key: str
    :ivar password: Device's unique password
    :vartype password: str
    :ivar actuator_references: Actuator references present on device
    :vartype actuator_references: List[str]
    """

    key: str
    password: str
    actuator_references: List[str] = field(default_factory=list)



"""Firmware handler for file installation and version reporting."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod


class FirmwareHandler(ABC):
    """Handle firmware installation and get current firmware version."""

    @abstractmethod
    def install_firmware(self, firmware_file_path: str) -> None:
        """
        Handle the installation of the firmware file.

        :param firmware_file_path: Path where the firmware file is located
        :type firmware_file_path: str
        """
        raise NotImplementedError()

    @abstractmethod
    def get_current_version(self) -> str:
        """
        Obtain device's current firmware version.

        :returns: version
        :rtpe: str
        """
        raise NotImplementedError()



"""OS File Management module."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import base64
import hashlib
import math
import os
import shutil
from tempfile import NamedTemporaryFile
from threading import Timer
from typing import Callable
from typing import IO
from typing import List
from typing import Optional
from urllib.parse import urlparse

import requests

from wolk import logger_factory
from wolk.interface.file_management import FileManagement
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.file_transfer_package import FileTransferPackage


class OSFileManagement(FileManagement):
    """
    File transfer manager.

    Enables device to transfer files from WolkAbout IoT Platform
    package by package or/and URL download as well as report list of files
    currently on device and delete them on request.
    """

    def __init__(
        self,
        status_callback: Callable[[str, FileManagementStatus], None],
        packet_request_callback: Callable[[str, int, int], None],
        url_status_callback: Callable[
            [str, FileManagementStatus, Optional[str]], None
        ],
    ) -> None:
        """
        Enable file management for device.

        :param status_callback: Reporting current file transfer status
        :type status_callback: Callable[[FileManagementStatus], None]
        :param packet_request_callback: Request file packet from Platform
        :type packet_request_callback: Callable[[str, int, int], None]
        :param url_status_callback: Report file url download status
        :type url_status_callback: Callable[[str, FileManagementStatus, Optional[str]], None]
        """
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )

        self.status_callback = status_callback
        self.packet_request_callback = packet_request_callback
        self.url_status_callback = url_status_callback
        self.preferred_package_size = 0
        self.max_file_size = 0
        self.file_directory = ""
        self.current_status: Optional[FileManagementStatus] = None
        self.max_retries = 3
        self.next_package_index: Optional[int] = None
        self.expected_number_of_packages: Optional[int] = None
        self.retry_count: Optional[int] = None
        self.request_timeout: Optional[Timer] = None
        self.last_package_hash = 32 * b"\x00"
        self.temp_file: Optional[IO[bytes]] = None
        self.file_name: Optional[str] = None
        self.file_size: Optional[int] = None
        self.file_hash: Optional[str] = None
        self.file_url: Optional[str] = None

    def configure(
        self,
        preferred_package_size: int,
        max_file_size: int,
        file_directory: str,
    ) -> None:
        """
        Configure options for file management module.

        :param preferred_package_size: Size in bytes
        :type preferred_package_size: int
        :param max_file_size: Maximum file size that can be stored
        :type max_file_size: int
        :param file_directory: Path to where files are stored
        :type file_directory: str
        """
        self.preferred_package_size = preferred_package_size
        self.max_file_size = max_file_size
        self.file_directory = file_directory

        if not os.path.exists(os.path.abspath(self.file_directory)):
            os.makedirs(os.path.abspath(self.file_directory))

    def handle_upload_initiation(
        self, file_name: str, file_size: int, file_hash: str
    ) -> None:
        """
        Start making package requests and set status to file transfer.

        :param file_name: File name
        :type file_name: str
        :param file_size: Size in bytes
        :type file_size: int
        :param file_hash: base64 encoded sha256 hash of file
        :type file_hash: str
        """
        if self.current_status is not None:
            self.logger.warning(
                "Not in idle state, ignoring file upload initiation"
            )
            return

        if self.preferred_package_size == 0:
            self.logger.warning("File management module not configured")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.TRANSFER_PROTOCOL_DISABLED,
            )
            self.status_callback(file_name, self.current_status)
            self.current_status = None
            return

        self.logger.info("Starting file transfer")
        self.logger.info(
            f"File name: {file_name} ; "
            f"File size: {file_size} ; "
            f"File hash: {file_hash}"
        )

        if file_size > self.max_file_size:
            self.logger.warning(
                f"File size {file_size} is greater than"
                f" maximum supported file size {self.max_file_size}"
            )
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.UNSUPPORTED_FILE_SIZE,
            )
            self.status_callback(file_name, self.current_status)
            self.current_status = None
            return

        self.expected_number_of_packages = math.ceil(
            file_size / self.preferred_package_size
        )

        if os.path.exists(
            os.path.join(os.path.abspath(self.file_directory), file_name)
        ):
            valid_file = False

            sha256_received_file_hash = base64.b64decode(
                file_hash + ("=" * (-len(file_hash) % 4))
            )
            sha256_file_hash = hashlib.sha256()

            existing_file = open(
                os.path.join(os.path.abspath(self.file_directory), file_name),
                "rb",
            )

            for chunk_index in range(self.expected_number_of_packages):

                existing_file.seek(chunk_index * self.preferred_package_size)
                chunk = existing_file.read(self.preferred_package_size)
                if not chunk:
                    self.logger.error("File size too small!")
                    break

                sha256_file_hash.update(chunk)

            sha256_file_hash = sha256_file_hash.digest()  # type: ignore
            valid_file = sha256_received_file_hash == sha256_file_hash

            if valid_file:
                self.logger.info(
                    "File requested for transfer already on device"
                )
                self.current_status = FileManagementStatus(
                    FileManagementStatusType.FILE_READY
                )
                self.status_callback(file_name, self.current_status)

                self.current_status = None
                self.last_package_hash = 32 * b"\x00"
                return

        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_TRANSFER
        )
        self.temp_file = NamedTemporaryFile(mode="a+b", delete=False)
        self.file_name = file_name
        self.file_size = file_size
        self.file_hash = file_hash
        self.next_package_index = 0
        self.retry_count = 0

        self.logger.info(f"Initializing file transfer: '{file_name}'")
        self.status_callback(self.file_name, self.current_status)

        if self.file_size < self.preferred_package_size:
            self.packet_request_callback(
                self.file_name, self.next_package_index, self.file_size + 64
            )
        else:
            self.packet_request_callback(
                self.file_name,
                self.next_package_index,
                self.preferred_package_size + 64,
            )

        self.request_timeout = Timer(60.0, self._timeout)
        self.request_timeout.start()

    def handle_file_upload_abort(self) -> None:
        """Abort file upload and revert to idle status."""
        self.logger.info("Aborting file upload.")
        if self.temp_file:
            self.temp_file.close()

        self.temp_file = None
        self.file_name = None
        self.file_size = None
        self.file_hash = None
        self.current_status = None
        self.next_package_index = None
        self.expected_number_of_packages = None
        self.retry_count = None
        self.request_timeout = None
        self.last_package_hash = 32 * b"\x00"

    def handle_file_binary_response(
        self, package: FileTransferPackage
    ) -> None:
        """
        Validate received package and store or use callback to request again.

        :param package: Package of file being transfered.
        :type package: FileTransferPackage
        """
        self.logger.debug(
            f"Previous hash: {package.previous_hash!r} ; "
            f"Data size: {len(package.data)} ; "
            f"Current hash: {package.current_hash!r}"
        )

        if self.current_status is None:
            self.logger.warning(
                "File transfer not in progress, ignoring package"
            )
            return

        if self.request_timeout:
            self.request_timeout.cancel()

        valid_package = False

        if (
            len(package.previous_hash)
            + len(package.data)
            + len(package.current_hash)
            < 65
        ):
            self.logger.error("Received package size too small!")
        else:
            data_hash = hashlib.sha256(package.data).digest()
            if package.current_hash != data_hash:
                self.logger.error(
                    f"Data hash '{data_hash!r}' does not match "
                    f"expected hash '{package.current_hash!r}' !"
                )
            else:
                self.logger.debug(
                    f"Valid package #{self.next_package_index} received"
                )
                valid_package = True

        if (
            not valid_package
            or self.last_package_hash != package.previous_hash
        ):
            self.logger.warning("Received invalid file package")
            self.retry_count += 1  # type: ignore

            if self.retry_count >= self.max_retries:
                self.logger.error(
                    "Retry count exceeded, aborting file transfer"
                )
                self.current_status = FileManagementStatus(
                    FileManagementStatusType.ERROR,
                    FileManagementErrorType.RETRY_COUNT_EXCEEDED,
                )
                self.status_callback(
                    self.file_name, self.current_status  # type: ignore
                )
                self.handle_file_upload_abort()
                return

            self.logger.info(
                f"Requesting package #{self.next_package_index} again"
            )
            self.packet_request_callback(
                self.file_name,  # type: ignore
                self.next_package_index,  # type: ignore
                self.preferred_package_size + 64,
            )

            self.request_timeout = Timer(60.0, self._timeout)
            self.request_timeout.start()
            return

        self.last_package_hash = package.current_hash

        try:
            self.temp_file.write(package.data)  # type: ignore
            self.temp_file.flush()  # type: ignore
            os.fsync(self.temp_file)  # type: ignore
        except Exception:
            self.logger.error(
                "Failed to write package, aborting file transfer"
            )
            self.current_status.status = FileManagementStatusType.ERROR
            self.current_status.error = (
                FileManagementErrorType.FILE_SYSTEM_ERROR
            )
            self.status_callback(
                self.file_name, self.current_status  # type: ignore
            )
            self.handle_file_upload_abort()
            return

        self.next_package_index += 1  # type: ignore

        if (
            self.next_package_index  # type: ignore
            < self.expected_number_of_packages
        ):
            self.logger.debug(
                f"Stored package, requesting "
                f"#{self.next_package_index}/"
                f"#{self.expected_number_of_packages}"
            )
            self.packet_request_callback(
                self.file_name,  # type: ignore
                self.next_package_index,
                self.preferred_package_size + 64,
            )

            self.request_timeout = Timer(60.0, self._timeout)
            self.request_timeout.start()
            return

        valid_file = False

        sha256_received_file_hash = base64.b64decode(
            self.file_hash + ("=" * (-len(self.file_hash) % 4))  # type: ignore
        )
        sha256_file_hash = hashlib.sha256()

        for index in range(self.expected_number_of_packages):  # type: ignore

            self.temp_file.seek(  # type: ignore
                index * self.preferred_package_size
            )
            chunk = self.temp_file.read(  # type: ignore
                self.preferred_package_size
            )
            if not chunk:
                self.logger.error("File size too small!")
                break

            sha256_file_hash.update(chunk)

        sha256_file_hash = sha256_file_hash.digest()  # type: ignore
        valid_file = sha256_received_file_hash == sha256_file_hash

        if not valid_file:
            self.logger.error("Invalid file - File hash does not match")
            self.current_status.status = FileManagementStatusType.ERROR
            self.current_status.error = (
                FileManagementErrorType.FILE_HASH_MISMATCH
            )
            self.status_callback(
                self.file_name, self.current_status  # type: ignore
            )
            self.handle_file_upload_abort()
            return

        if not os.path.exists(os.path.abspath(self.file_directory)):
            os.makedirs(os.path.abspath(self.file_directory))

        file_path = os.path.join(
            os.path.abspath(self.file_directory),
            self.file_name,  # type: ignore
        )

        shutil.copy2(
            os.path.realpath(self.temp_file.name), file_path  # type: ignore
        )
        self.temp_file.close()  # type: ignore

        if not os.path.exists(file_path):
            self.logger.error(f"File failed to store to at: {file_path}")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.FILE_SYSTEM_ERROR,
            )
            self.status_callback(
                self.file_name, self.current_status  # type: ignore
            )
            self.handle_file_upload_abort()
            return

        self.logger.info(f"Received file '{self.file_name}'")
        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_READY
        )
        self.status_callback(
            self.file_name, self.current_status  # type: ignore
        )

        self.current_status = None
        self.retry_count = None
        self.last_package_hash = 32 * b"\x00"

    def handle_file_url_download_initiation(self, file_url: str) -> None:
        """
        Start file transfer from specified URL.

        :param file_url: URL from where to download file
        :type file_url: str
        """
        if self.current_status is not None:
            self.logger.warning(
                "Not in idle state, ignoring file upload initiation"
            )
            return

        if not bool(urlparse(file_url).scheme):
            self.logger.error(f"Received URL '{file_url}' is not valid!")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.MALFORMED_URL,
            )
            self.url_status_callback(file_url, self.current_status, None)
            self.handle_file_upload_abort()
            return

        self.file_url = file_url
        self.file_name = self.file_url.split("/")[-1]
        file_path = os.path.join(
            os.path.abspath(self.file_directory), self.file_name
        )
        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_TRANSFER
        )
        self.url_status_callback(file_url, self.current_status, self.file_name)

        response = requests.get(file_url)
        with open(file_path, "ab") as file:
            file.write(response.content)
            file.flush()
            os.fsync(file)  # type: ignore

        if not os.path.exists(file_path):
            self.logger.error(f"File failed to store to at: {file_path}")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.FILE_SYSTEM_ERROR,
            )
            self.url_status_callback(
                file_url, self.current_status, self.file_name
            )
            self.handle_file_url_download_abort()
            return

        self.logger.info(f"File obtained from URL: '{file_url}'")
        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_READY
        )
        self.url_status_callback(file_url, self.current_status, self.file_name)

        self.current_status = None

    def handle_file_url_download_abort(self) -> None:
        """Abort file URL download."""
        self.logger.info("Aborting URL download")
        self.file_url = None
        self.file_name = None
        self.current_status = None

    def get_file_list(self) -> List[str]:
        """
        Return list of files present on device.

        :returns: file_list
        :rtype: List[str]
        """
        file_list = os.listdir(os.path.abspath(self.file_directory))

        if file_list:
            for item in file_list:
                if not os.path.isfile(
                    os.path.join(os.path.abspath(self.file_directory), item)
                ) or item.startswith("."):
                    file_list.remove(item)
        else:
            file_list = []

        self.logger.debug(f"Files on device: {file_list}")
        return file_list

    def get_file_path(self, file_name: str) -> Optional[str]:
        """
        Return path to file if it exists.

        :param file_name: File for which to get path
        :type file_name: str
        :returns: file_path
        :rtype: Optional[str]
        """
        self.logger.debug(f"Get file path for file: {file_name}")
        file_path = None
        for file in self.get_file_list():
            if file == file_name:
                file_path = os.path.join(
                    os.path.abspath(self.file_directory), file_name
                )
                break
        self.logger.debug(f"File path: {file_path}")
        return file_path

    def handle_file_list_confirm(self) -> None:
        """Acknowledge file list response from WolkAbout IoT Platform."""
        ...

    def handle_file_delete(self, file_name: str) -> None:
        """
        Delete file from device.

        :param file_name: File to be deleted
        :type file_name: str
        """
        self.logger.info(f"Attempting to delete file: '{file_name}'")
        file_path = os.path.join(
            os.path.abspath(self.file_directory), file_name
        )

        if os.path.exists(file_path):
            os.remove(file_path)

    def handle_file_purge(self) -> None:
        """Delete all files from device."""
        for file in os.listdir(os.path.abspath(self.file_directory)):
            if not os.path.isfile(
                os.path.join(os.path.abspath(self.file_directory), file)
            ) or file.startswith("."):
                continue
            os.remove(os.path.join(os.path.abspath(self.file_directory), file))

    def _timeout(self) -> None:
        """Revert to idle state when timeout occurs."""
        self.logger.error("Timed out waiting for next package, aborting")
        self.current_status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.UNSPECIFIED_ERROR,
        )
        self.status_callback(
            self.file_name, self.current_status  # type: ignore
        )
        self.handle_file_upload_abort()
